(function (){
    "use strict";
})();

const axios = require("axios");

/**
 * @typedef {"EX"|"PX"|"EXAT"|"PXAT"|"NX"|"XX"|"KEEPTTL"|"GET"} ExpireMode
 */

class Redis{
    /**
     * Create a redis operator instance, Just on you request source, that will connect.
     * 
     * If connect error, It will throw error, else return a redis instance
     * @param {Object} params
     * @param {String} params.host - the connect host string
     * @param {(Number|String)=} params.port - the connect port
     * @param {{
     *      protocol:String,
     *      resourceURL:String,
     *      remote:boolean,
     *      proxy:boolean
     *      security:boolean
     * }=} params.options
     * @property {String} params.options.protocol - "http" | "https" | "".., the protocol for communication
     * @property {String} params.options.resourceURL - specify a source URL if you need
     * @property {Boolean} params.options.remote - specify the redis server is on local or on remote, by defalut, it's true
     * @property {Boolean} params.options.proxy - specify the redis server status, proxy that mean redis server not same host with proxy server
     * @property {Boolean} params.options.security - specify the communication protocol is it with http or http-security, by default ,it's true
     * 
     * @example simple connect with https
     * //for example, if connect the default product
     * const redis = new Redis({host:"redis.avancevl.com"});
     * redis.set("hello","world");
     * redis.get("hello").then((result) => {
     *      console.log(result);//world
     * }).catch((error) => {
     *      console.log(error);//if has error
     * });
     */
    constructor({host, port=0, options={protocol:"http", resourceURL:"", remote:true, proxy:true, security:true }}) {
        this.host = host;
        this.port = parseInt(port+10)-10;
        this.options = options;
        if (this.options.security === true) {
            this.options.protocol = "https";
        }
        this.baseUrl = (() => {
            if (this.options.resourceURL !== "") {
                return this.options.resourceURL;
            }
            let header = this.options.protocol;
            let host = this.host;
            let port = this.port;
            return `${header}://${host}${parseInt(port) === 0?"":":"+port}`;
        })();
        Redis.connectTest({host,port,options});
    }

    /**
     * Check connect status, if can't connect. will throw error
     * @async
     * @param {Object} params
     * @param {String} params.host
     * @param {Number|String} params.port
     * @param {{
     *      protocol:"String"|"http"|"https",
     *      resourceURL:"String"
     * }=} params.options
     * @returns {Promise<void>}
     */
    static async connectTest({host, port, options={protocol:"https", resourceURL:""}}){
        if (options.resourceURL === "") {
            options.resourceURL = `${options.protocol}://${host}:${port}`;
        }
        if (parseInt(port) === 0) {
            options.resourceURL = `${options.protocol}://${host}`;
        }
        let response = await axios.get(`${options.resourceURL.split("/proxy")[0]}/healthCheck`);
        if (response.status !== 200) {
            throw new Error("Can't connect to redis service");
        }
    }

    /**
     * Send a request to redis serivce, then return the payload
     * @param {String} path 
     * @param {Object|JSON|String} payload 
     * @returns {Promise<any|Error>}
     */
    async sendRequest(path, payload) {
        try {
            let _resourceURL = `${this.baseUrl}${this.options.proxy?"/proxy":""}${path}`;
            const { data } = await axios.post(_resourceURL, payload);
            try {
                return JSON.parse(data);
            } catch (error) {
                return data;
            }
        } catch (error) {
            console.log(error);
            return new Error(error);
        }
    }
    /**
     * Get the set value of the key
     * @param {String} key 
     * @returns {Promise<any|Error>}
     * 
     * @example strings
     * 
     * redis.set("hello","world");
     * 
     * redis.get("hello").then((result) => {
     *      console.log(result);//"world"
     * }).catch((error) => {
     *      console.log(error);
     * });
     * 
     * @example objects
     * redis.set("hello",{ a: 1, b: 'two', c: [ 1, 2, 3 ] });
     * 
     * redis.get("hello").then((result) => {
     *      console.log(result);//{ a: 1, b: 'two', c: [ 1, 2, 3 ] } :object
     * }).catch((error) => {
     *      console.log(error);
     * });
     */
    async get(key) {
        return this.sendRequest("/get", {key:key});
    }

    /**
     * Set value to redis service
     * @param {String} key
     * @param {JSON|String} value 
     * @param {(String|ExpireMode)=} expireMode - expireMode unit, EX = seconed
     * @param {(Number|String)=} time - the expire value
     * @returns {Promise<"OK"|Error>} - if success will return "OK"
     * @example strings
     * 
     * redis.set("hello","world").then((result) => {
     *      console.log(result);// "OK"
     * }).catch((error) => {
     *      console.log(error);//error handle
     * });
     * 
     * @example objects
     * 
     * redis.set("hello",{ a: 1, b: 'two', c: [ 1, 2, 3 ] }).then((result) => {
     *      console.log(result);// "OK"
     * }).catch((error) => {
     *      console.log(error);//error handle
     * });
     */
    async set(key, value, expireMode, time) {
        if (typeof(value) !== "string") {
            value = JSON.stringify(value);
        }
        if (expireMode && time) {
            return this.sendRequest("/setWithTime", {key:key, value:value, expireMode:expireMode, time:time});
        }
        return this.sendRequest("/set", {key:key, value:value});
    }

    /**
     * multi get
     * @param {Array<String>} keys - Array for keys, it will enumerate the array and get datas
     * @return {Promise<Array<String>|Error>}
     */
    async mget(keys) {
        if (Array.isArray(keys)) {
            keys = JSON.stringify(keys);
        } else {
            throw new Error("type Error");
        }
        return this.sendRequest("/getMulti", {key:keys});
    }

    /**
     * multi set, it use pipeline to accomplish
     * @param {{
     *      [key:string]: string
     * }|Array<{
     *      key:String,
     *      value:JSON|String,
     *      expireMode:ExpireMode|String,
     *      time:Number|String
     * }>} datasets 
     * @returns {Promise<"OK"|[Error, "OK"]|Error>}
     * 
     * @example mapping
     * 
     * let dataset = {
     *      "a":"1",
     *      "b":"2",
     *      "c":"[1,2,3]",
     *      "d":'{"A":1,"B":2}'
     * }
     * 
     * @example time_limit
     * 
     * let dataset = [
     *      {
     *          key:String,
     *          value:String|JSON,
     *          expireMode:"EX"|"PX"|"EXAT"|"PXAT"|"NX"|"XX"|"KEEPTTL"|"GET",
     *          time:Number
     *      }
     * ]
     */
    async mset(datasets) {
        if (Array.isArray(datasets)) {
            return this.sendRequest("/setMultiWithTime", {dataSet:JSON.stringify(datasets)});
        }
        return this.sendRequest("/setMulti", {dataSet:JSON.stringify(datasets)});
    }

    /**
     * operator redis list data, it like array.
     * 
     * the method will push the data to specify list
     * @param {String} key - the list key name
     * @param {Array<any>|JSON} value - the value who you want to push
     * @returns {Promise<JSON|Number|Error>} - the number for data length
     * 
     * @example
     * 
     * redis.lpush("hello_list",["1", "2", 3, 4, 5]).then((result) => {
     *      console.log(result);//5 --the number of the length
     * }).catch((error) => {
     *      console.log(error);
     * });
     * 
     * redis.lrange("hello_list", 0, -1).then((result) => {
     *      console.log(result);//["1", "2", 3, 4, 5]
     * }).catch((error) => {
     *      console.log(error);//error handle
     * });
     */
    async lpush(key, value) {
        if (Array.isArray(value)) {
            value = JSON.stringify(value);
        }
        return this.sendRequest("/lpush",{key:key, value:value});
    }

    /**
     * get list data given a range
     * @param {String} key - list key name
     * @param {Number} start - index start
     * @param {Number} end - index end
     * @returns {Promise<Array<any>|Error>}
     * 
     * @example
     * 
     * redis.lpush("hello_list",["1", "2", 3, 4, 5]).then((result) => {
     *      console.log(result);//5 --the number of the length
     * }).catch((error) => {
     *      console.log(error);
     * });
     * 
     * redis.lrange("hello_list", 0, -1).then((result) => {
     *      console.log(result);//["1", "2", 3, 4, 5]
     * }).catch((error) => {
     *      console.log(error);//error handle
     * });
     */
    async lrange(key, start, end) {
        return this.sendRequest("/lrange", {key, start, end});
    }

    /**
     * remove the data equal the value, remove {count} amount of the data
     * 
     * If count > 0, from head to foot
     * If count < 0, form foot to head
     * If count = 0, remove all data equal value in the list
     * @param {String} key 
     * @param {Number} count 
     * @param {String|JSON} value 
     * @returns {Promise<Number|Error>}
     */
    async lrem(key, count, value) {
        if (typeof(value) !== "string") {
            value = JSON.stringify(value);
        }
        return this.sendRequest("/lrem", {key, count, value});
    }

    /**
     * pop a data for you specify list
     * @param {String} key -The key of list
     * @returns {Promise<JSON|Error>}
     */
    async lpop(key) {
        return this.sendRequest("/lpop", {key});
    }

    /**
     * trim a specify list, that mean just reserve the in range data
     * @param {String} key - list key name
     * @param {Number} start - index start
     * @param {Number} end - index end
     * @returns {Promise<"OK"|Error>}
     * 
     * @example
     * 
     * //hello_list = [1, 2, 3, 4, 5];
     * 
     * redis.ltrim("hello_list", 0, 1).then((result) => {
     *      console.log(result);//"OK"
     *      //hello_list = [1, 2]
     * }).catch((error) => {
     *      console.log(error);
     * });
     */
    async ltrim(key, start, end) {
        return this.sendRequest("/ltrim", {key, start, end});
    }

    /**
     * set the data specify index to the list
     * @param {String} key - list key name
     * @param {Number} index - list index
     * @param {JSON|String} value - change the index value
     * @returns {Promise<"OK"|Error>}
     * @example
     * 
     * //hello_list: ["1", 2, 3, "4"]
     * 
     * redis.lset("hello_list", 0, 5).then((result) => {
     *      console.log(result);//"OK"
     *      //hello_list: [5, 2, 3, "4"]
     * }).catch((error) => {
     *      console.log(error);
     * });
     */
    async lset(key, index, value) {
        return this.sendRequest("/lset", {key, index, value});
    }

    /**
     * pipeline for exec redis command, pipeline will help redis performance
     * @param {Array<Array<String>>|String[][]|JSON} commands 
     * @returns {Promise<[Error,any][]|Error>}
     * 
     * @example
     * let commands = [
     *      ["set","key","value"],
     *      ["get","key"],
     *      ["lpush","key","value"]
     * ]
     * 
     * redis.pipeline(commands).then((result) => {
     *      console.log(result);
     *      //["OK","value","OK"]
     * }).catch((error) => {
     *      console.log(error);
     * });
     */
    async pipeline(commands) {
        if (typeof(commands) !== "string") {
            commands = JSON.stringify(commands);
        }
        return this.sendRequest("/pipeline", {commands});
    }

    /**
     * get the keys in redis server that it has.
     * @param {String|RegExp} patterns - regexp
     * @returns {Promise<String[]|Error>} - string array for you want query keys
     * @example all
     * //suppose to have:
     * //hello - "world"
     * //teddy - "1565"
     * //teddy1565 - "ssss"
     * 
     * 
     * redis.keys("*").then((result) => {
     *      console.log(result);
     *      //[ "hello", "teddy", "teddy1565" ]
     * }).catch((error) => {
     *      console.log(error);
     * });
     * 
     * @example undefined
     * 
     * redis.keys("teddy15651").then((result) => {
     *      console.log(result); //[] - an empty array
     * }).catch((error) => {
     *      console.log(error);
     * });
     */
    async keys(patterns) {
        return this.sendRequest("/keys", {patterns});
    }

    /**
     * query the specify key-value ttl (remaining survival time)
     * @param {String | Array<String>} key 
     * @returns {Promise<Number|Error>}
     */
    async ttl(key) {
        if (Array.isArray(key)) {
            return this.sendRequest("/ttlMulti", { key:JSON.stringify(key) });
        }
        return this.sendRequest("/ttl", { key });
    }

    /**
     * delete specify dataset of the key(s)
     * @param {String|Array<String>} key 
     * @returns {Promise<"OK"|String<Number>|Error>}
     */
    async del(key) {
        if (Array.isArray(key)) {
            key = JSON.stringify(key);
            return this.sendRequest("/delMulti", {key});
        }
        return this.sendRequest("/del", {key});
    }
}

module.exports = Redis;
